'
' Maze creator and solver
' Create a maze and display it on the serial terminal, then solve it and display it.
'
' Written by Les Johnson for the Proton compiler
'
    Include "Amicus18.inc"              ' Configure the compiler for a PIC18F25K20 at 64MHz. i.e. An Amicus18 board
'
'--------------------------------------------------------------------------

$define cColumns 24                                 ' The amount of columns in the maze
$define cRows 21                                    ' The amount of rows in the maze
$define cCellAmount $eval ((cRows * cColumns) / 2)  ' The size of the array used for the maze

$define cCOLSdiv2 $eval (cColumns / 2)
$define cROWSdiv2 $eval (cRows / 2)

$define cWALL_NORTH    %00000001
$define cWALL_WEST     %00000010

$define cWALL_MASK     %00000011
$define cWALL_MASK_INV %11111100

$define cDIR_NORTH     %00000000
$define cDIR_WEST      %00000100

$define cDIR_SOUTH     %00001000
$define cDIR_EAST      %00001100

$define cDIR_MASK      %00001100
$define cDIR_MASK_INV  %11110011
$define cDIR_NONE      %11111111
'
' Create variables
'
    Dim bMazeFlags As Byte System
    Dim tNotVisited_Result As bMazeFlags.0
    Dim tWallTest_Result As bMazeFlags.1

    Dim Row_bCellOffsetGet As Byte
    Dim Col_bCellOffsetGet As Byte
    Dim Cell_OffsetGet_bResult As Byte
    Dim Row_bCellGet As Byte
    Dim Col_bCellGet As Byte
    Dim bValue_CellGet As Byte
    Dim bOffset_CellGet As Byte
    Dim Cell_Get_bResult As Byte
    Dim Row_bCellSet As Byte
    Dim Col_bCellSet As Byte
    Dim Wall_Get_bResult As Byte
    Dim Dir_Get_bResult As Wall_Get_bResult
    Dim Value_bCellSet As Wall_Get_bResult
    Dim Row_bWallTest As Byte
    Dim bOffset_CellSet As Row_bWallTest
    Dim Col_bWallTest As Byte
    Dim bMask_CellSet As Col_bWallTest
    Dim Wall_bWallTest As Byte
    Dim bRow_MazeDraw As Byte
    Dim bCol_MazeDraw As Byte
    Dim Row_bWallSet As Byte
    Dim Col_bWallSet As Byte
    Dim Wall_bWallSet As Byte
    Dim Direction_DirSet As Wall_bWallSet
    
    Dim Row_bWallClear As Byte
    Dim Col_bWallClear As Byte
    Dim Wall_bWallClear As Byte


    Dim Maze_bDirection As Byte
    Dim Maze_bRow As Byte
    Dim Maze_bColumn As Byte

    Dim Maze_wCellsRemaining As Word

    Dim Maze_bNeighbourCount As Byte
    Dim Row_bNotVisited As Row_bWallTest
    Dim Col_bNotVisited As Col_bWallTest

    Dim Maze_bNeighbour[4] As Byte
    Dim Maze_bCellsArray[cCellAmount] As Byte Heap      ' Create an array to hold the maze

$ifndef True
    $define True 1
$endif
$ifndef False
    $define False 0
$endif

'-----------------------------------------------------------------------------------------------
    GoTo Main                                           ' Jump over the subroutines
'-----------------------------------------------------------------------------------------------
' Determine the cell offset given a Row and Column
' Input     : Row_bCellOffsetGet holds the row
'           : Col_bCellOffsetGet holds the column
' Output    : WREG
' Notes     : None
'
$define Cell_Offset_Get(pRow, pCol, pResult) '
    Row_bCellOffsetGet = pRow            '
    Col_bCellOffsetGet = pCol            '
    GoSub _CellOffsetGet                  '
    pResult = WREG

_CellOffsetGet:
    WREG = (Row_bCellOffsetGet * cCOLSdiv2) + (Col_bCellOffsetGet / 2)
    Return

'-----------------------------------------------------------------------------------------------
' Set a cell
' Input     : Row_bCellSet holds the row
'           : Col_bCellSet holds the column
' Output    : None
' Notes     : Set a cell to Value
'
$define Cell_Set(pRow, pCol, pValue) '
    Row_bCellSet = pRow      '
    Col_bCellSet = pCol      '
    Value_bCellSet = pValue  '
    GoSub _CellSet

_CellSet:
    Cell_Offset_Get(Row_bCellSet, Col_bCellSet, bOffset_CellSet)

    bMask_CellSet = $F0                                 ' Clear the upper nibble
    If Col_bCellSet.0 = 0 Then
        bMask_CellSet = $0F                             ' Clear the lower nibble
        Value_bCellSet = Value_bCellSet << 4          ' Move int the lower nibble
    EndIf
    Maze_bCellsArray[bOffset_CellSet] = (Maze_bCellsArray[bOffset_CellSet] & bMask_CellSet) | Value_bCellSet
    Return

'-----------------------------------------------------------------------------------------------
' Return the value of a cell
' Input     : Row_bCellGet holds the row
'           : Col_bCellGet holds the column
' Output    : WREG holds the type of cell
' Notes     : None
'
$define Cell_Get(pRow, pCol, pResult) '
    Row_bCellGet = pRow     '
    Col_bCellGet = pCol     '
    GoSub _CellGet           '
    pResult = WREG

_CellGet:
    bValue_CellGet = $0F
    If Row_bCellGet < cRows Then If Col_bCellGet < cColumns Then
        Cell_Offset_Get(Row_bCellGet, Col_bCellGet, bOffset_CellGet)
        bValue_CellGet = Maze_bCellsArray[bOffset_CellGet]
        If Col_bCellGet.0 = 0 Then
            bValue_CellGet = bValue_CellGet >> 4
        EndIf
    EndIf
    WREG = bValue_CellGet & $0F
    Return

'-----------------------------------------------------------------------------------------------
' Create a wall
' Input     : Row_bWallSet holds the row to set
'           : Col_bWallSet holds the column to set
'           : Wall_bWallSet holds the wall type
' Output    : None
' Notes     : None
'
$define Wall_Set(pRow, pCol, pWall) '
    Row_bWallSet = pRow   '
    Col_bWallSet = pCol   '
    Wall_bWallSet = pWall '
    GoSub _WallSet

_WallSet:
    Cell_Get(Row_bWallSet, Col_bWallSet, Cell_Get_bResult)
    Cell_Set(Row_bWallSet, Col_bWallSet, (Cell_Get_bResult & cWALL_MASK_INV) | Wall_bWallSet)
    Return

'-----------------------------------------------------------------------------------------------
' Return the value of a cell
' Input     : Row_bWallGet holds the row
'           : Col_bWallGet holds the column
' Output    : WREG holds the type of cell
' Notes     : None
'
$define Wall_Get(pRow, pCol, pResult) '
    Row_bCellSet = pRow '
    Col_bCellSet = pCol '
    GoSub _WallGet       '
    pResult = WREG

_WallGet:
    Cell_Get(Row_bCellSet, Col_bCellSet, Cell_Get_bResult)
    WREG = Cell_Get_bResult & cWALL_MASK
    Return

'-----------------------------------------------------------------------------------------------
' Clear a wall
' Input     : Row_bWallClear holds the row
'           : Col_bWallClear holds the column
'           : Wall_bWallClear holds the type of wall to clear
' Output    : None
' Notes     :
'
$define Wall_Clear(pRow, pCol, pWall) '
    Row_bWallClear = pRow   '
    Col_bWallClear = pCol   '
    Wall_bWallClear = pWall '
    GoSub _WallClear

_WallClear:
    Wall_Get(Row_bWallClear, Col_bWallClear, Wall_Get_bResult)
    Wall_Set(Row_bWallClear, Col_bWallClear, Wall_Get_bResult & ~Wall_bWallClear)
    Return

'-----------------------------------------------------------------------------------------------
' Input     :
' Output    : tWallTest_Result
' Notes     :
'
$define Wall_Test(pRow, pCol, pWall, pResult) '
    Row_bWallTest = pRow     '
    Col_bWallTest = pCol     '
    Wall_bWallTest = pWall   '
    GoSub _WallTest           '
    pResult = tWallTest_Result

_WallTest:
    tWallTest_Result = False
    Wall_Get(Row_bWallTest, Col_bWallTest, Wall_Get_bResult)
    Wall_Get_bResult = Wall_Get_bResult & Wall_bWallTest
    If Wall_Get_bResult <> 0 Then
        tWallTest_Result = True
    EndIf
    Return

'-----------------------------------------------------------------------------------------------
' Input     :
' Output    : None
' Notes     :
'
$define Direction_Set(pRow, pCol, pDirection) '
    Row_bWallSet = pRow           '
    Col_bWallSet = pCol           '
    Direction_DirSet = pDirection '
    GoSub _DirSet

_DirSet:
    Cell_Get(Row_bWallSet, Col_bWallSet, Cell_Get_bResult)
    Cell_Set(Row_bWallSet, Col_bWallSet, (Cell_Get_bResult & cDIR_MASK_INV) | Direction_DirSet)
    Return

'-----------------------------------------------------------------------------------------------
' Input     :
' Output    : WREG
' Notes     :
'
$define Direction_Get(pRow, pCol, pResult) '
    Row_bCellSet = pRow '
    Col_bCellSet = pCol '
    GoSub _DirGet        '
    pResult = WREG

_DirGet:
    Cell_Get(Row_bCellSet, Col_bCellSet, WREG)
    WREG = WREG & cDIR_MASK
    Return

'-----------------------------------------------------------------------------------------------
' Input     :
' Output    : tNotVisited_Result
' Notes     :
'
$define Not_Visited(pRow, pCol) '
    Row_bNotVisited = pRow  '
    Col_bNotVisited = pCol  '
    GoSub _NotVisited

_NotVisited:
    tNotVisited_Result = False
    Wall_Get(Row_bNotVisited, Col_bNotVisited, Wall_Get_bResult)
    If Wall_Get_bResult = (cWALL_NORTH | cWALL_WEST) Then
        Wall_Get(Row_bNotVisited, Col_bNotVisited + 1, Wall_Get_bResult)
        Wall_Get_bResult = Wall_Get_bResult & cWALL_WEST
        If Wall_Get_bResult <> 0 Then
            Wall_Get(Row_bNotVisited + 1, Col_bNotVisited, Wall_Get_bResult)
            Wall_Get_bResult = Wall_Get_bResult & cWALL_NORTH
            If Wall_Get_bResult <> 0 Then
                tNotVisited_Result = True
            EndIf
        EndIf
    EndIf
    Return

'-----------------------------------------------------------------------------------------------
' Maze_Draw
' Input     : None
' Output    : None
' Notes     : None
'
$define Maze_Draw() GoSub _Maze_Draw

_Maze_Draw:
    bRow_MazeDraw = 0                                                       ' \
    Repeat                                                                  ' / Create a loop For the amount of rows in the maze
        bCol_MazeDraw = 0                                                   ' \
        Repeat                                                              ' / Create a loop For the amount of columns in the maze
            Wall_Test(bRow_MazeDraw, bCol_MazeDraw, cWALL_NORTH, tWallTest_Result) ' Check for a north wall
            If tWallTest_Result = True Then                                 ' Is there a north wall?
                If bRow_MazeDraw = 0 And bCol_MazeDraw = 0 Then             ' Yes. So are we at the entrance to the maze?
                    HRSOut "+  "                                            ' Yes. So do not create a wall
                Else                                                        ' Otherwise...
                    HRSOut "+--"                                            ' Create a wall
                EndIf
            Else                                                            ' Otherwise...
                Direction_Get(bRow_MazeDraw, bCol_MazeDraw, Dir_Get_bResult)
                If Dir_Get_bResult = 0 Then
                    HRSOut "+  "
                Else
                    Direction_Get(bRow_MazeDraw - 1, bCol_MazeDraw, Dir_Get_bResult)
                    If Dir_Get_bResult <> 0 Then
                        HRSOut "+.."
                    Else
                        HRSOut "+  "
                    EndIf
                EndIf
            EndIf
            Inc bCol_MazeDraw
        Until bCol_MazeDraw >= cColumns
        HRSOut "+", 13                                              ' Create an east joint wall

        bCol_MazeDraw = 0
        Repeat
            Wall_Test(bRow_MazeDraw, bCol_MazeDraw, cWALL_WEST, tWallTest_Result)
            If tWallTest_Result = True Then
                Direction_Get(bRow_MazeDraw, bCol_MazeDraw, Dir_Get_bResult)
                If Dir_Get_bResult <> 0 Then
                    HRSOut "|.."
                Else
                    HRSOut "|  "
                EndIf
            Else
                Direction_Get(bRow_MazeDraw, bCol_MazeDraw, Dir_Get_bResult)
                If Dir_Get_bResult <> 0 Then
                    Direction_Get(bRow_MazeDraw, bCol_MazeDraw - 1, Dir_Get_bResult)
                    If Dir_Get_bResult = 0 Then
                        HRSOut " .."
                    Else
                        HRSOut "..."
                    EndIf
                Else
                    HRSOut "   "
                EndIf
            EndIf
            Inc bCol_MazeDraw
        Until bCol_MazeDraw >= cColumns
        If bRow_MazeDraw = cRows - 1 And bCol_MazeDraw >= cColumns Then    ' Have we reached the exit from the maze?
            HRSOut " Exit", 13                                          ' Yes. So create an open exit
        Else                                                            ' Otherwise...
            HRSOut "|", 13                                              ' Create an east perimeter wall
        EndIf
        Inc bRow_MazeDraw
    Until bRow_MazeDraw >= cRows
'
' Draw the bottom wall of the maze
'
    bCol_MazeDraw = 0
    Repeat
        HRSOut "+--"
        Inc bCol_MazeDraw
    Until bCol_MazeDraw >= cColumns
    HRSOut "+"
    Return
    
'-----------------------------------------------------------------------------------------------
' Maze_Create
' Input     : None
' Output    : None
' Notes     : None
'
$define Maze_Create() GoSub _Maze_Create

_Maze_Create:
    Maze_wCellsRemaining = 0
    Repeat
        Maze_bCellsArray[Maze_wCellsRemaining] = ((cWALL_NORTH | cWALL_WEST) << 4) | (cWALL_NORTH | cWALL_WEST)
        Inc Maze_wCellsRemaining
    Until Maze_wCellsRemaining >= ((cRows * cColumns) / 2)

    Maze_wCellsRemaining = (cRows * cColumns) - 1
    Maze_bRow = 0
    Maze_bColumn = 0

    While Maze_wCellsRemaining <> 0
        Maze_bNeighbourCount = 0

        If Maze_bRow <> 0 Then
            Not_Visited(Maze_bRow - 1, Maze_bColumn)
            If tNotVisited_Result = True Then
                Maze_bNeighbour[0] = cDIR_NORTH
                Inc Maze_bNeighbourCount
            EndIf
        EndIf

        If Maze_bColumn <> 0 Then
            Not_Visited(Maze_bRow, Maze_bColumn - 1)
            If tNotVisited_Result = True Then
                Maze_bNeighbour[Maze_bNeighbourCount] = cDIR_WEST
                Inc Maze_bNeighbourCount
            EndIf
        EndIf

        If Maze_bRow < (cRows - 1) Then
            Not_Visited(Maze_bRow + 1, Maze_bColumn)
            If tNotVisited_Result = True Then
                Maze_bNeighbour[Maze_bNeighbourCount] = cDIR_SOUTH
                Inc Maze_bNeighbourCount
            EndIf
        EndIf

        If Maze_bColumn < (cColumns - 1) Then
            Not_Visited(Maze_bRow, Maze_bColumn + 1)
            If tNotVisited_Result = True Then
                Maze_bNeighbour[Maze_bNeighbourCount] = cDIR_EAST
                Inc Maze_bNeighbourCount
            EndIf
        EndIf

        If Maze_bNeighbourCount <> 0 Then
            Maze_bDirection = Maze_bNeighbour[(Random) // Maze_bNeighbourCount]

            Select Maze_bDirection
                Case cDIR_NORTH
                    Wall_Clear(Maze_bRow, Maze_bColumn, cWALL_NORTH)
                    Direction_Set(Maze_bRow - 1, Maze_bColumn, cDIR_SOUTH)
                Case cDIR_WEST
                    Wall_Clear(Maze_bRow, Maze_bColumn, cWALL_WEST)
                    Direction_Set(Maze_bRow, Maze_bColumn - 1, cDIR_EAST)
                Case cDIR_SOUTH
                    Inc Maze_bRow
                    Wall_Clear(Maze_bRow, Maze_bColumn, cWALL_NORTH)
                    Direction_Set(Maze_bRow, Maze_bColumn, cDIR_NORTH)
                    Maze_bDirection = cDIR_NONE
                Case cDIR_EAST
                    Inc Maze_bColumn
                    Wall_Clear(Maze_bRow, Maze_bColumn, cWALL_WEST)
                    Direction_Set(Maze_bRow, Maze_bColumn, cDIR_WEST)
                    Maze_bDirection = cDIR_NONE
            EndSelect
            Dec Maze_wCellsRemaining
        Else
            Direction_Get(Maze_bRow, Maze_bColumn, Maze_bDirection)
        EndIf
        Select Maze_bDirection
            Case cDIR_NORTH
                Dec Maze_bRow
            Case cDIR_WEST
                Dec Maze_bColumn
            Case cDIR_SOUTH
                Inc Maze_bRow
            Case cDIR_EAST
                Inc Maze_bColumn
        EndSelect
    Wend

    Maze_bRow = 0
    Repeat
        Maze_bColumn = 0
        Repeat
            Direction_Set(Maze_bRow, Maze_bColumn, 0)
            Inc Maze_bColumn
        Until Maze_bColumn >= cColumns
        Inc Maze_bRow
    Until Maze_bRow >= cRows
    Return

'-----------------------------------------------------------------------------------------------
' Maze_Solve
' Input     : None
' Output    : None
' Notes     : None
'
$define Maze_Solve() GoSub _Maze_Solve

_Maze_Solve:
    Maze_bDirection = cDIR_EAST
    Direction_Set(0, 0, cDIR_MASK)
    Maze_bRow = 0
    Maze_bColumn = 0

    While Maze_bRow <> (cRows - 1) Or Maze_bColumn <> (cColumns - 1)
        Select Maze_bDirection
            Case cDIR_NORTH
                Wall_Test(Maze_bRow, Maze_bColumn, cWALL_NORTH, tWallTest_Result)
                If tWallTest_Result = False Then
                    Direction_Get(Maze_bRow - 1, Maze_bColumn, Dir_Get_bResult)
                    If Dir_Get_bResult <> 0 Then
                        Direction_Set(Maze_bRow, Maze_bColumn, 0)
                    EndIf
                    Dec Maze_bRow
                    Maze_bDirection = cDIR_WEST
                Else
                    Maze_bDirection = cDIR_EAST
                EndIf

            Case cDIR_WEST
                Wall_Test(Maze_bRow, Maze_bColumn, cWALL_WEST, tWallTest_Result)
                If tWallTest_Result = False Then
                    Direction_Get(Maze_bRow, Maze_bColumn - 1, Dir_Get_bResult)
                    If Dir_Get_bResult <> 0 Then
                        Direction_Set(Maze_bRow, Maze_bColumn, 0)
                    EndIf
                    Dec Maze_bColumn
                    Maze_bDirection = cDIR_SOUTH
                Else
                    Maze_bDirection = cDIR_NORTH
                EndIf

            Case cDIR_SOUTH
                Wall_Test(Maze_bRow + 1, Maze_bColumn, cWALL_NORTH, tWallTest_Result)
                If tWallTest_Result = False Then
                    Direction_Get(Maze_bRow + 1, Maze_bColumn, Dir_Get_bResult)
                    If Dir_Get_bResult <> 0 Then
                        Direction_Set(Maze_bRow, Maze_bColumn, 0)
                    EndIf
                    Inc Maze_bRow
                    Maze_bDirection = cDIR_EAST
                Else
                    Maze_bDirection = cDIR_WEST
                EndIf

            Case cDIR_EAST
                Wall_Test(Maze_bRow, Maze_bColumn + 1, cWALL_WEST, tWallTest_Result)
                If tWallTest_Result = False Then
                    Direction_Get(Maze_bRow, Maze_bColumn + 1, Dir_Get_bResult)
                    If Dir_Get_bResult <> 0 Then
                        Direction_Set(Maze_bRow, Maze_bColumn, 0)
                    EndIf
                    Inc Maze_bColumn
                    Maze_bDirection = cDIR_NORTH
                Else
                    Maze_bDirection = cDIR_SOUTH
                EndIf
        EndSelect
        Direction_Set(Maze_bRow, Maze_bColumn, cDIR_MASK)
    Wend
    Return

'-----------------------------------------------------------------------------------------------
' Start of the main program
'
Main:
    Seed $0235
    Maze_Create()
    HRSOut "Entrance\r"

    Maze_Draw()

    HRSOut "\r\rSolved\r\r"
    Maze_Solve()
    Maze_Draw()

    
